"""Give users some control over the server and request lifecycles.
"""
from aspen import is_callable


class Hooks(dict):
    """Model a collection of application extension points.

    I want attribute access, but dict.__repr__ and __eq__.

    """

    def __init__(self, names):
        """Takes a list of hook names.
        """
        if isinstance(names, basestring):
            raise TypeError("Please pass in an iterable of unicodes.")
        self.__names = names
        for name in names:
            hook = Hook()
            try:
                setattr(self, name, hook)
            except NameError:
                raise ValueError("Hooks names must be valid Python attribute "
                                 "names.")
            dict.__setitem__(self, name, hook)

    def __getitem__(self, name):
        raise NotImplementedError("Please use attribute access.")

    def __setitem__(self, name, value):
        raise NotImplementedError("Please use attribute access.")


class Hook(list):
    """Model a single point where callbacks can be registered.
    """

    def append(self, obj):
        raise NotImplementedError("Please use register.")

    def register(self, obj):
        """Extend to ensure that obj is callable. Could check signature, too.
        """
        if not is_callable(obj):
            raise TypeError("Hooks must be callable objects; this isn't: %s." 
                            % str(obj))
        list.append(self, obj)

    def run(self, thing):
        """Takes a request/response/website.
        """
        for hook in self:
            thing = hook(thing) or thing
        return thing



class Context(dict):
    """Model the execution context for a Resource.
    """

    def __init__(self, request):
        """Takes a Request object.
        """
        self.website    = None # set in dynamic_resource.py
        self.body       = request.body
        self.headers    = request.headers
        self.cookie     = request.headers.cookie
        self.path       = request.line.uri.path
        self.qs         = request.line.uri.querystring
        self.request    = request
        self.socket     = None
        self.channel    = None
        self.context    = self

        # http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html
        for method in ['OPTIONS', 'GET', 'HEAD', 'POST', 'PUT', 'DELETE', 
                       'TRACE', 'CONNECT']:
            self[method] = (method == request.line.method)
            setattr(self, method, self[method])

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError("")

    def __setattr__(self, name, value):
        self[name] = value



"""This is a simple tool to restart a process when it dies.

It's designed to restart aspen in development when it dies because files
have changed and you set changes_reload to 'yes'.

    http://aspen.io/thrash/

"""

def foo():

    # Import in here to capture KeyboardInterrupt during imports.
    import os
    import subprocess
    import sys
    import time

    # set unbuffered - http://stackoverflow.com/a/181654/253309
    sys.stdout = os.fdopen(sys.stdout.fileno(), 'wb', 0)

    if len(sys.argv) < 2:
        print ("usage: %s <child> [child opts and args]" 
               % os.path.basename(sys.argv[0]))
        sys.exit(1)

    BACKOFF_MIN = 0.10  # Start looping here.
    BACKOFF_MAX = 3.20  # Throttle back to here.
    PROMPT_AFTER = 60   # Give the user this much time to fix the error.
    INITIAL_WAIT = 15   # Give the user this much time to read the error.

    n = 0 
    backoff = BACKOFF_MIN
    cumulative_time = 0

    while 1:

        # The child process exited.
        # ========================= 
        # Log restart attempts after the initial launch.

        n += 1
        backoff = min(backoff * 2, BACKOFF_MAX)
        if n > 1:
            m = "---- Restart #%s " % n
            print
            print m + ('-' * (79-len(m)))


        # Execute the child process.
        # ==========================
        # Then wait for it to return, dealing with INT.

        proc = subprocess.Popen( sys.argv[1:]
                               , stdout=sys.stdout
                               , stderr=sys.stderr
                                )
        try:
            status = proc.wait()
        except KeyboardInterrupt:
            status = proc.wait()

        if status == 75:
            print ("Received INT in child.")


        # Decide how to proceed.
        # ======================

        if n == 1:
            # This is the first time we've thrashed. Give the user time to 
            # parse the (presumed) traceback.
            cumulative_time += INITIAL_WAIT 
            try:
                time.sleep(INITIAL_WAIT)
            except KeyboardInterrupt:
                # Allow user to fast-track this step.
                
                # reset
                n = 0
                backoff = BACKOFF_MIN
                cumulative_time = 0

        elif cumulative_time < PROMPT_AFTER:      
            # We've given the user time to parse the traceback. Now thrash 
            # for a while.
            cumulative_time += backoff
            time.sleep(backoff)
            
        else:
            # We've been thrashing for a while. Pause.
            print
            try:
                raw_input("Press any key to start thrashing again. ")
            except KeyboardInterrupt:
                print

            # reset
            n = 0
            backoff = BACKOFF_MIN
            cumulative_time = 0


def main():
    try:
        foo()
    except KeyboardInterrupt:
        import time
        time.sleep(0.1)  # give child stdio time to flush
        print "Received INT in thrash, exiting."
